#!/usr//env python

import clip, json, logging, logtool, termcolor, yaml
from addict import Dict

LOG = logging.getLogger (__name__)
COLOUR_ERROR = "red"
COLOUR_DEBUG = "blue"
COLOUR_INFO = "white"
COLOUR_INFO_BAD = "magenta"
COLOUR_WARN = "yellow"
COLOUR_DATA = "green"
COLOUR_REPORT = "green"
COLOUR_FIELDNAME = "cyan"
COLOUR_VALUE = "green"

class CmdIO (object):

  @logtool.log_call
  def __init__ (self, conf = None):
    self.conf = conf if conf else Dict ()

  @logtool.log_call (log_args = False, log_rc = False)
  def _safe_echo (self, msg, nl):
    if isinstance (msg, str):
      msg = unicode (msg, encoding = "utf8")
    return clip.echo (msg, nl = nl)

  @logtool.log_call (log_args = False, log_rc = False)
  def colourise (self, msg, colour):
    return (termcolor.colored (msg, colour) if not self.conf.nocolour
            else msg)

  @logtool.log_call (log_args = False, log_rc = False)
  def debug (self, msg, nl = True):
    if not self.conf.quiet and self.conf.verbose:
      self._safe_echo (self.colourise (msg, COLOUR_DEBUG), nl = nl)

  @logtool.log_call (log_args = False, log_rc = False)
  def info (self, msg, err = False, nl = True):
    if not self.conf.quiet:
      self._safe_echo (self.colourise (
        "INFO: " + msg, COLOUR_INFO if not err else COLOUR_INFO_BAD),
                       nl = nl)

  @logtool.log_call (log_args = False, log_rc = False)
  def error (self, msg, nl = True):
    if not self.conf.quiet:
      self._safe_echo (self.colourise ("ERROR: " + msg, COLOUR_ERROR),
                       nl = nl)

  @logtool.log_call (log_args = False, log_rc = False)
  def warn (self, msg, nl = True):
    if not self.conf.quiet:
      self._safe_echo (self.colourise ("WARNING: " + msg, COLOUR_WARN),
                       nl = nl)

  @logtool.log_call (log_args = False, log_rc = False)
  def report (self, msg, err = False, nl = True):
    if not self.conf.quiet:
      self._safe_echo (self.colourise (
          msg, COLOUR_REPORT if not err else COLOUR_ERROR), nl = nl)

  @logtool.log_call (log_args = False, log_rc = False)
  def data (self, msg, err = False, nl = True):
    if not self.conf.quiet:
      if self.conf.json:
        s = json.dumps (msg, indent = 2)
      else:
        s = yaml.dump (msg, width = 70, indent = 2,
                       default_flow_style = False)
      self._safe_echo (self.colourise (
          s, COLOUR_DATA if not err else COLOUR_ERROR), nl = nl)



#! /usr/bin/env python

from __future__ import absolute_import
import clip, logtool, logging, sys
from addict import Dict
from encodings.utf_8 import StreamWriter
from functools import partial
from path import Path
from .cmdio import CmdIO
from .cmd import Action

from ._version import get_versions
__version__ = get_versions ()['version']
del get_versions

logging.basicConfig (level = logging.WARN)
LOG = logging.getLogger (__name__)
TIME_T, TIME_STR = logtool.now ()

# Python 2 will try to coerce its output stream to match the terminal it is
# printing to. If we pipe the output, it will attempt to do plain ASCII
# encoding, which breaks on unicode characters. The following will change
# the default encoding for pipes from ascii to utf-8.
#   See : https://wiki.python.org/moin/PrintFails
sys.stdout = StreamWriter(sys.stdout)
sys.stderr = StreamWriter(sys.stderr)

APP = clip.App (name = "s3clumper")

CONFIG = Dict ({
  "check": False,
  "force": False,
  "nocolour": False,
  "delete": False,
  "quiet": False,
  "time_str": TIME_STR,
  "time_t": TIME_T,
  "verbose": False,
})
IO = CmdIO (conf = CONFIG)

@logtool.log_call
def option_setopt (option, value):
  CONFIG[option] = value

@logtool.log_call
def option_version (opt): # pylint: disable = W0613
  print __version__
  clip.exit ()

@logtool.log_call
def option_logging (flag):
  logging.root.setLevel (logging.DEBUG)
  CONFIG.debug = flag

@APP.main (name = Path (sys.argv[0]).basename (),
           description = "Aggregate S3 prefixes into tarballs in S3",
           tree_view = "-H")
@clip.flag ("-H", "--HELP",
            help = "Help for all sub-commands")
@clip.flag ("-c", "--check", name = "check",
            help = "Don't check for target (may over-write)",
            callback = partial (option_setopt, "check"))
@clip.flag ("-C", "--nocolour", name = "nocolour",
            help = "Suppress colours in reports",
            callback = partial (option_setopt, "nocolour"))
@clip.flag ("-D", "--debug", name = "debug", help = "Enable debug logging",
            callback = option_logging)
@clip.flag ("-d", "--delete", name = "delete",
            help = "Don't delete source files",
            callback = partial (option_setopt, "delete"))
@clip.flag ("-q", "--quiet", name = "quiet",
            help = "Be quiet, be vewy vewy quiet",
            callback = partial (option_setopt, "verbose"))
@clip.flag ("-v", "--verbose", name = "verbose",
            help = "Verbose output",
            callback = option_version)
@clip.flag ("-V", "--version", name = "verbose",
            help = "Report installed version",
            callback = option_version)
@clip.flag ("-z", "--compress", name = "compress",
            help = "Don't compress the target",
            callback = partial (option_setopt, "compress"))
@clip.arg ("from", help = "S3 URL prefix to clump")
@clip.arg ("to", help = "S3 URL for target clump")
@logtool.log_call
def app_main (**kwargs):
  if not CONFIG.conf.debug:
    logging.basicConfig (level = logging.ERROR)
  if not sys.stdout.isatty ():
    option_setopt ("nocolour", True)
  CONFIG.url_from = kwargs["from"]
  CONFIG.url_to = kwargs["to"]
  Action (CONFIG).run ()

@logtool.log_call
def main ():
  try:
    APP.run ()
  except KeyboardInterrupt:
    pass
  except clip.ClipExit as e:
    sys.exit (e.status)
  except Exception as e:
    logtool.log_fault (e)
    sys.exit (1)

if __name__ == "__main__":
  main ()



#!/usr/bin/env python

from __future__ import absolute_import
import boto3, botocore, clip, logging, logtool, os, progressbar
import retryp, StringIO, tarfile, tempfile, threading, urlparse
from collections import namedtuple
from . import cmdio

LOG = logging.getLogger (__name__)

class _ProgressPercentage (object):

  # pylint: disable=too-few-public-methods

  @logtool.log_call
  def __init__ (self, filename, quiet):
    self._filename = filename
    self._size = float (os.path.getsize (filename))
    self._lock = threading.Lock ()
    self.quiet = quiet
    self.progress = (progressbar.ProgressBar (max_value = self._size)
                     if not quiet else None)

  @logtool.log_call
  def __enter__ (self):
    return self

  @logtool.log_call
  def __exit__ (self, *args):
    if not self.quiet:
      self.progress.finish ()

  @logtool.log_call
  def __call__ (self, bytes_amount):
    if self.progress:
      with self._lock:
        self.progress.update (bytes_amount)

class Action (cmdio.CmdIO):

  @logtool.log_call
  def __init__ (self, args):
    cmdio.CmdIO.__init__ (self, conf = args)
    self.args = args
    self.p_from = self._parse_url ("Source", args.url_from)
    self.p_to = self._parse_url ("Destination", args.url_to)
    self.compress = args.compress
    self.check = args.check
    self.s3 = boto3.resource ("s3")
    self.keys = None

  @logtool.log_call
  def _parse_url (self, typ, url):
    _urlspec = namedtuple ("_UrlSpec", ["protocol", "bucket", "key"])
    p = urlparse.urlparse (url)
    rc = _urlspec (protocol = p.scheme, bucket = p.netloc,
                   key = p.path[1:] if p.path.startswith ("/") else p.path)
    if rc.protocol != "s3":
      self.error ("%s protocol is not s3: %s" % (typ, url))
      clip.exit (err = True)
    return rc

  @retryp.retryp (expose_last_exc = True, log_faults = True)
  @logtool.log_call
  def _cleanup (self):
    if not self.args.delete:
      if not self.args.quiet:
        print "Deleting..."
      for key in (progressbar.ProgressBar () (self.keys)
                  if not self.args.quiet else self.keys):
        key.delete ()

  @retryp.retryp (expose_last_exc = True, log_faults = True)
  @logtool.log_call
  def _send (self, out_f):
    client = boto3.client('s3')
    if not self.args.quiet:
      print "Sending..."
    with _ProgressPercentage (out_f.name, self.args.quiet) as cb:
      client.upload_file (
        out_f.name, self.p_to.bucket, self.p_to.key, Callback = cb)

  @retryp.retryp (expose_last_exc = True, log_faults = True)
  @logtool.log_call
  def _get_file (self, client, bucket, key, fhandle):
    client.download_fileobj (bucket, key, fhandle)
    fhandle.flush ()

  @logtool.log_call
  def _entar (self, out_f):
    mode = "w" if self.compress else "w:gz"
    client = boto3.client('s3')
    with tarfile.open (out_f.name, mode = mode) as tar:
      if not self.args.quiet:
        print "Collecting..."
      for key in (progressbar.ProgressBar () (self.keys)
                  if not self.args.quiet else self.keys):
        with tempfile.NamedTemporaryFile (prefix = "s3clumper_tf__") as f:
          self._get_file (client, self.p_from.bucket, key.key, f)
          tar.add (f.name, arcname = key.key)
    out_f.flush ()
    out_f.seek (0)

  @retryp.retryp (expose_last_exc = True, log_faults = True)
  @logtool.log_call
  def _list_from (self):
    bucket = self.s3.Bucket (self.p_from.bucket)
    self.keys = [k for k in bucket.objects.filter (
      Prefix = self.p_from.key)]

  @retryp.retryp (expose_last_exc = True, log_faults = True)
  @logtool.log_call
  def _check (self):
    if self.check:
      return False
    try:
      boto3.client ("s3").head_object (
        Bucket = self.p_to.bucket,
        Key = self.p_to.key)
      self.error ("Target exists: %s" % self.args.url_to)
      return True
    except botocore.exceptions.ClientError as e:
      if int (e.response['Error']['Code']) == 404:
        client = boto3.client('s3')
        client.upload_fileobj (
          StringIO.StringIO (), self.p_to.bucket, self.p_to.key)
        return False
      raise

  @logtool.log_call
  def run (self):
    if self._check ():
      clip.exit (err = True)
    self._list_from ()
    with tempfile.NamedTemporaryFile (prefix = "s3clumper_tar__") as out_f:
      self._entar (out_f)
      self._send (out_f)
      self._cleanup ()
      if not self.args.quiet:
        print ""



#! /usr/bin/env python

from setuptools import setup, find_packages
import versioneer
from io import open

setup (
    name = "s3clumper",
    version = versioneer.get_version (),
    description = "Aggregate an S3 folder into a file in another",
    long_description = open ("README.rst", "r", encoding = "utf-8").read (),
    cmdclass = versioneer.get_cmdclass (),
    classifiers = [],
    keywords = "S3 aggregate tarball compress",
    author = "J C Lawrence",
    author_email = "claw@kanga.nu",
    url = "https://pypi.python.org/pypi/s3clumper",
    license = "LGPL v3.0",
    packages = find_packages (exclude = ["tests",]),
    package_data = {
    },
    zip_safe = True,
    install_requires = [line.strip ()
                        for line in open ("requirements.txt", "r",
                                          encoding = "utf-8").readlines ()],
    entry_points = {
        "console_scripts": [
            "s3clumper = s3clumper.main:main",
        ],
    },
)
